# -*- coding: utf-8 -*-
"""Device class module for Hekr API"""
import asyncio
import logging
import re

from aiohttp import ClientSession, WSMsgType, client_exceptions
from typing import Tuple, Union, Optional, AnyStr, Any
from enum import IntEnum, Enum
from json import dumps, loads
from random import getrandbits
from base64 import b64encode

from .types import CommandSearchIdentifier
from .aioudp import open_datagram_endpoint, open_remote_endpoint
from .command import Command
from .exceptions import (
    DeviceProtocolNotSetException,
    HeartbeatFailedException,
    LocalAuthenticationFailedException,
    CloudAuthenticationFailedException,
    CommandFailedException,
    DeviceConnectionMissingException,
    AuthenticationFailedException
)
from .const import (
    DEFAULT_APPLICATION_ID,
    DEFAULT_DEVICE_PORT,
    DEFAULT_REQUEST_RETRIES,
    DEFAULT_RETRY_DELAY
)

_LOGGER = logging.getLogger(__name__)

SENSITIVE_INFO_REGEX = re.compile(r'("(ctrlKey|token)"\s*:\s*")[^"]+(")')
SENSITIVE_INFO_FILTER = lambda content: SENSITIVE_INFO_REGEX.sub(r'\1<redacted>\3', content)

def device_id_from_mac_address(mac_address: Union[str, bytearray]) -> str:
    """Convert mac address to device ID

    Arguments:
        mac_address {Union[str, bytearray]} -- MAC-address string (dashed format)

    Returns:
        str -- Device ID
    """
    if isinstance(mac_address, bytearray):
        mac_address = mac_address.hex()

    for delimiter in [':', '-', ' ']:
        mac_address = mac_address.replace(delimiter, '')

    mac_address = mac_address.upper()
    return 'ESP_2M_' + mac_address

class DeviceResponseState(Enum):
    """Response state of `make_request`"""
    SUCCESS = 0
    FAILURE = 1
    WAIT_NEXT = 2

class DeviceConnectionType(Enum):
    """Connection types for devices"""
    NONE = 0
    LOCAL = 1
    CLOUD = 2

class Device:
    """Device class for Hekr API"""

    def __init__(self,
                 device_id: str,
                 control_key: str,
                 host: Union[str, type(None)]=None,
                 port: int = DEFAULT_DEVICE_PORT,
                 protocol: Optional['Protocol']=None,
                 application_id: str = DEFAULT_APPLICATION_ID
                 ):
        self.device_id = device_id
        self.local_address = (host, port)
        self.protocol = protocol
        self.application_id = application_id

        self.__control_key = control_key

        self.__last_frame_number = 0
        self.__last_message_id = 0

        self.__cloud_endpoint = None
        self.__cloud_authenticated = False
        self.__cloud_token = None
        self.__cloud_domain = None

        self.__local_endpoint = None
        self.__local_authenticated = False

    def __str__(self) -> str:
        """Generates a string representation of the device

        Returns:
            str -- String representation (human-distinguishable)
        """
        if self.available_connection_type == DeviceConnectionType.LOCAL:
            return '{}@{}:{}'.format(self.device_id, *self.local_address)
        elif self.available_connection_type == DeviceConnectionType.CLOUD:
            return '{}@hekr-cloud'.format(self.device_id)
        elif self.available_connection_type == DeviceConnectionType.NONE:
            return '{}'.format(self.device_id)

    def __repr__(self) -> str:
        """Generates a debug representation of the device

        Returns:
            str -- Debug representation (python-like)
        """
        return '<{}({}, {})>'.format(
            self.__class__.__name__,
            self.device_id,
            ':'.join(map(str,self.local_address))
                if self.available_connection_type == DeviceConnectionType.LOCAL
                else 'hekr-cloud'
                if self.available_connection_type == DeviceConnectionType.CLOUD
                else DeviceConnectionType.NONE.name
        )

    def generate_request(self,
                         action:str,
                         params:dict=None,
                         message_id:int=None,
                         connection_type:DeviceConnectionType=None
                         ) -> bytes:
        """Generate request string (JSON format)

        Arguments:
            action {str} -- Action name (read README.md for discovered actions list)

        Keyword Arguments:
            params {dict} -- parameters array for actions (not required for some actions) (default: {None})
            message_id {int} -- message id (default: last message id for set connection)
            connection_type {DeviceConnectionType} -- set connection (default: {self.available_connection})

        Raises:
            DeviceConnectionMissingException: Device connection missing

        Returns:
            bytes -- Bytes-encoded JSON request
        """
        connection_type = connection_type or self.available_connection_type

        if message_id is None:
            self.__last_message_id += 1
            message_id = self.__last_message_id

        request_dict = {
            "msgId": message_id,
            "action": action,
        }

        if action != 'heartbeat':
            request_dict["params"] = {
                "ctrlKey": self.__control_key,
                "devTid": self.device_id
            }

            if connection_type == DeviceConnectionType.LOCAL:
                # No additional data required for local connections
                pass
            elif connection_type == DeviceConnectionType.CLOUD:
                request_dict["params"]["appTid"] = self.application_id

        if isinstance(params, dict):
            request_dict['params'].update(params)

        return (message_id, str.encode(dumps(request_dict)))

    @property
    def available_connection_type(self) -> DeviceConnectionType:
        """Retrieve best available connection type for communication.

        Best connection type is determined by the availability of the sockets.
        If none are open, this property will return `DeviceConnectionType.NONE`.

        Returns:
            DeviceConnectionType -- Available connection type
        """
        # @TODO: check for more stuff to determine connection type
        if self.__local_endpoint is not None:
            return DeviceConnectionType.LOCAL
        if self.__cloud_endpoint is not None:
            return DeviceConnectionType.CLOUD

        return DeviceConnectionType.NONE

    def set_control_key(self, control_key: str):
        """Sets device control key

        Arguments:
            control_key {str} -- Control key
        """
        self.__control_key = control_key

    def set_cloud_settings(self, cloud_token: str, cloud_domain: str) -> None:
        """Sets settings for cloud authentication

        Arguments:
            cloud_token {str} -- Cloud authentication token
            cloud_domain {str} -- Cloud authentication domain
        """
        if self.__cloud_domain != cloud_domain or self.__cloud_token != cloud_token:
            self.__cloud_authenticated = False

        self.__cloud_token = cloud_token
        self.__cloud_domain = cloud_domain

    async def open_socket_local(self) -> bool:
        """Opens local socket to device

        Returns:
            bool -- State of socket opening
        """
        _LOGGER.debug('Opening local endpoint on device %s', self)
        self.__local_endpoint = await open_remote_endpoint(*self.local_address)
        return True

    def _generate_websocket_key(self) -> str:
        """Generates WebSocket secret key to send as a header with initial request.

        Returns:
            str -- WebSocket secret key
        """
        raw_key = bytes(getrandbits(8) for _ in range(16))
        return b64encode(raw_key).decode()

    async def open_socket_cloud(self) -> bool:
        """Opens cloud socket to device

        Returns:
            bool -- State of socket opening
        """
        _LOGGER.debug('Opening cloud endpoint on device %s', self)
        try:
            session = ClientSession()
            self.__cloud_endpoint = await session.ws_connect(
                'https://{}:186/'.format(self.__cloud_domain),
                headers={
                    'Sec-WebSocket-Key': self._generate_websocket_key,
                    'Sec-WebSocket-Version': '13',
                    'Connection': 'upgrade',
                    'Upgrade': 'websocket'
                })
            _LOGGER.debug('Cloud endpoint opened on device %s', self)
            return True
        except client_exceptions.ClientConnectionError:
            _LOGGER.exception('Client connection could not be established')
            await session.close()
            return False

    def _handle_incoming_response(self,
                                  response:AnyStr,
                                  message_id:int
                                  ) -> Tuple[DeviceResponseState, Any]:
        """Handle incoming response packet (decode from existing string)

        Arguments:
            response {AnyStr} -- Response string
            message_id {int} -- Message ID (used for compliance verification)

        Returns:
            Tuple[DeviceResponseState, Any] -- Tuple of (state, data)
        """
        if isinstance(response, bytes):
            data = response.decode('utf-8')
        else:
            data = response
        data = data.strip()

        _LOGGER.debug('Received response for device %s: %s', self, SENSITIVE_INFO_FILTER(data))

        response_dict = loads(data)

        # @TODO: compare message ids to verify correct request sequence
        #response_message_id = response_dict.get('msgId')

        if response_dict.get('action') == 'appDevAuthResp':
            """Handle authentication responses"""
            if response_dict.get('code') == 200:
                _LOGGER.debug('Authentication executed succesfully on device %s', self)
                self.__local_authenticated = True
                state = DeviceResponseState.SUCCESS
            else:
                _LOGGER.error('Authentication failed on device %s', self)
                self.__local_authenticated = False
                state = DeviceResponseState.FAILURE

        elif response_dict.get('action') == 'appLoginResp':
            if response_dict.get('code') == 200:
                _LOGGER.debug('Authentication executed succesfully on device %s', self)
                self.__cloud_authenticated = True
                state = DeviceResponseState.SUCCESS
            else:
                _LOGGER.error('Authentication failed on device %s', self)
                self.__cloud_authenticated = False
                state = DeviceResponseState.FAILURE

        elif response_dict.get('action') == 'heartbeatResp':
            """Handle heartbeat responses"""
            if response_dict.get('code') == 200:
                _LOGGER.debug('Heartbeat executed succesfully on device %s', self)
                state = DeviceResponseState.SUCCESS
            else:
                _LOGGER.error('Heartbeat failed on device %s', self)
                state = DeviceResponseState.FAILURE

        elif response_dict.get('action') == 'appSendResp':
            """Handle responses from command requests"""
            if response_dict.get('code') == 200:
                (command, _, _) = self.protocol.decode(raw=response_dict['params']['data']['raw'])
                if command.response_command_id:
                    next_command = self.protocol.get_command(command.response_command_id)
                    _LOGGER.debug('Command %s should wait for next response from command %s on device %s', command, next_command, self)
                    state = DeviceResponseState.WAIT_NEXT
                else:
                    _LOGGER.debug('Command %s executed succesfully on device %s', command, self)
                    state = DeviceResponseState.SUCCESS
            else:
                _LOGGER.debug('Command %s failed on device %s', command, self)
                state = DeviceResponseState.FAILURE

        elif response_dict.get('action') == 'devSend':
            """Handle device status updates"""
            if not self.protocol:
                raise DeviceProtocolNotSetException(device=self)

            decoded = self.protocol.decode(raw=response_dict['params']['data']['raw'])

            _LOGGER.debug('Received command response for device %s: %s', self, decoded)

            state = DeviceResponseState.SUCCESS
            data = decoded
        else:
            _LOGGER.debug('Unknown response received on device %s: %s', self, response_dict)
            state = DeviceResponseState.FAILURE

        return (state, data)

    async def make_request(self, action: str, params: dict = None,
                     message_id: int = None,
                     frame_number: int = None,
                     connection_type: DeviceConnectionType = None):
        """Make device request

        Arguments:
            action {str} -- Action string

        Keyword Arguments:
            params {dict} -- Request parameters (default: {None})
            message_id {int} -- Message ID (default: {None})
            frame_number {int} -- Frame number (default: {None})
            connection_type {DeviceConnectionType} -- Connection type (default: {None})
            retries {int} -- Amount of attempts the request should be made after it fails

        Returns:
            dict -- Request response
        """
        connection_type = connection_type or self.available_connection_type
        if connection_type == DeviceConnectionType.NONE:
            raise DeviceConnectionMissingException(device=self)

        (message_id, request) = self.generate_request(
            action=action,
            params=params,
            connection_type=connection_type
        )

        # @TODO: maybe move encoding to `generate_request`?
        request_str = request.decode('utf-8')

        _LOGGER.debug('Composed request for device %s, content: %s', self, SENSITIVE_INFO_FILTER(request_str))

        if connection_type == DeviceConnectionType.LOCAL:
            if not self.__local_endpoint:
                await self.open_socket_local()
            self.__local_endpoint.send(request)
        elif connection_type == DeviceConnectionType.CLOUD:
            if not self.__cloud_endpoint:
                await self.open_socket_cloud()
            await self.__cloud_endpoint.send_str(request_str)

        retries = 0
        while retries < DEFAULT_REQUEST_RETRIES:
            if connection_type == DeviceConnectionType.LOCAL:
                data = await self.__local_endpoint.receive()
            elif connection_type == DeviceConnectionType.CLOUD:
                message = await self.__cloud_endpoint.receive()

                if message.type == WSMsgType.TEXT:
                    data = message.data
                elif message.type == WSMsgType.BINARY:
                    data = message.data.decode('utf-8')
                else:
                    return (DeviceResponseState.FAILURE, data)

            (state, data) = self._handle_incoming_response(data, message_id)

            if state == DeviceResponseState.WAIT_NEXT:
                retries += 1
                continue

            return (state, data)

        _LOGGER.debug('Received too many %s responses, marking request result as invalid on device %s', DeviceResponseState.WAIT_NEXT.name, self)
        return (DeviceResponseState.FAILURE, data)

    async def heartbeat(self,
                        connection_type: DeviceConnectionType = None):
        """Send heartbeat message

        Keyword Arguments:
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Raises:
            HeartbeatFailedException: Heartbeat message sending failed
        """
        _LOGGER.debug('Requesting heartbeat on device %s', self)
        return await self.make_request(
            action='heartbeat',
            params=False,
            connection_type=connection_type)

    async def authenticate(self, connection_type: DeviceConnectionType = None):
        """Authenticate with the device

        Keyword Arguments:
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Raises:
            LocalAuthenticationFailedException: Local authentication failed
            AuthenticationFailedException: Authentication failed
        """
        connection_type = connection_type or self.available_connection_type
        if connection_type == DeviceConnectionType.NONE:
            raise DeviceConnectionMissingException(device=self)

        if connection_type == DeviceConnectionType.LOCAL:
            return await self.make_request(
                    action='appDevAuth',
                    connection_type=connection_type)

        if connection_type == DeviceConnectionType.CLOUD:
            params = {"token": self.__cloud_token}
            return await self.make_request(
                    action='appLogin',
                    connection_type=connection_type,
                    params=params)

    async def command(self,
                command: Union[CommandSearchIdentifier, Command],
                data: dict = None,
                frame_number:int=None,
                return_decoded:bool=True,
                connection_type:DeviceConnectionType=None):
        """Execute device command and return response

        Arguments:
            command {Union[int, str, Command]} -- Command ID/name/object to use

        Keyword Arguments:
            data {dict} -- Data values for datagram (default: {None})
            return_decoded {bool} -- Extract and decode datagram from response (default: {True})
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Raises:
            DeviceProtocolNotSetException: Device does not have a protocol set
            CommandFailedException: Command failed to execute due to device code error
            CommandFailedException: Command failed to execute due to missing datagram
            CommandFailedException: Command failed to execute due to socket timeout

        Returns:
            dict -- decoded datagram values or full response dictionary
        """
        if not isinstance(command, Command):
            if self.protocol:
                command = self.protocol.get_command(command)
            else:
                raise DeviceProtocolNotSetException(self)

        if frame_number is None:
            self.__last_frame_number += 1
            frame_number = self.__last_frame_number

        connection_type = connection_type or self.available_connection_type

        raw = self.protocol.encode(
            data=data,
            command=command,
            frame_number=frame_number
        )

        request_dict = {"appTid": self.application_id, "data": {"raw": raw}}

        return await self.make_request(
            action='appSend',
            params=request_dict,
            connection_type=connection_type
        )


# -*- coding: utf-8 -*-
"""Command class module for HekrAPI"""

from enum import IntEnum, Enum
from typing import Union

from .argument import Argument
from .exceptions import HekrTypeError, HekrValueError

class FrameType(IntEnum):
    """Datagram frame types (per Hekr documentation)"""
    RECEIVE = 0x01
    SEND = 0x02
    DEVICE = 0xFE
    ERROR = 0xFF

class Command:
    """Command class for HekrAPI"""

    def __init__(self,
                 command_id: int,
                 frame_type: FrameType=FrameType.SEND,
                 name: str=None,
                 arguments:list=None,
                 response_command_id:int=None):
        """Command class constructor

        Arguments:
            command_id {int} -- Command identifier for datagrams
            name {str} -- Command name for calls
            frame_type {FrameType} -- Command frame type (per Hekr docs)
        """

        if isinstance(frame_type, str):
            try:
                frame_type = FrameType[frame_type.upper()]
            except KeyError:
                raise ValueError('`frame_type` must be one of the following values (case-insensitive) when passed as string: %s (got %s)' % (', '.join(['`'+value.name.lower()+'`' for value in FrameType]), '`'+frame_type+'`'))

        # circular checking prevents following private attributes
        # from being created automatically
        self.__response_command_id = None
        self.__frame_type = None

        self.command_id = command_id
        self.name = name
        self.response_command_id = response_command_id
        self.frame_type = frame_type
        self.arguments = arguments
        if arguments:
            self.arguments.extend(arguments)

    def __repr__(self) -> str:
        """Friendly command string conversion

        Returns:
            str -- Friendly debug representation
        """
        return '<{}("{}", {}, {})>'.format(
            self.__class__.__name__,
            self.name,
            self.command_id,
            self.frame_type.name
        )

    def __str__(self) -> str:
        """Return name of the command

        Returns:
            str -- Command name
        """
        return self.name

    def __int__(self) -> int:
        """Return command identifier

        Returns:
            int -- Command identifier
        """
        return self.command_id

    @property
    def command_id(self) -> int:
        """Getter for numeric command identifier"""
        return self.__command_id

    @command_id.setter
    def command_id(self, value:int):
        """Setter for numeric command identifier

        Checks whether `command_id` being set to a correct integer value
        between 0 and 255 (a boundary imposed by datagram protocol)

        Arguments:
            value {int} -- Numeric command identifier for given command

        Raises:
            HekrTypeError: Raised when the value fed is not an integer.
            HekrValueError: Raised when a value is out of bounds of bytes constraint.
        """
        if not isinstance(value, int):
            raise HekrTypeError(variable='command_id', expected=int, got=type(value))
        elif 255 < value < 0:
            raise HekrValueError(variable='command_id', expected='integer from 0 to 255', got=value)
        self.__command_id = value

    @property
    def frame_type(self) -> FrameType:
        """Getter for `frame_type` attribute"""
        return self.__frame_type

    @frame_type.setter
    def frame_type(self, value:FrameType):
        """Setter for `frame_type` attribute

        Coerces any incoming value on invocation to `FrameType` class instances.
        Throws exceptions when incorrect data is being fed.

        Arguments:
            value {Any} -- Value (coercible to `FrameType`)

        Raises:
            HekrTypeError: Raised when the value is not coercible to a `FrameType` instance
            HekrValueError: Raised when the value is not `FrameType.SEND` when `response_command_id` attribute is set.
        """
        try:
            if value is None:
                new_frame_type = FrameType.SEND
            elif isinstance(value, str):
                new_frame_type = FrameType[value.upper()]
            else:
                new_frame_type = FrameType(value)
        except KeyError:
            raise HekrValueError(variable='frame_type', expected=', '.join(map(lambda x: "'"+x.name.lower()+"'", FrameType))+" (case-insensitive)", got="'"+value+"'")
        except ValueError:
            raise HekrTypeError(variable='frame_type', expected='coercible to `%s`' % FrameType.__name__, got=type(value))

        if new_frame_type != FrameType.SEND and self.response_command_id is not None:
            raise HekrValueError(variable='frame_type', expected='`%s` due to `response_command_id` being set' % FrameType.SEND, got=new_frame_type)

        self.__frame_type = new_frame_type

    @property
    def arguments(self) -> list:
        return self.__arguments

    @arguments.setter
    def arguments(self, value):
        if isinstance(value, list):
            invalid_types = [type(x) for x in value
                if not isinstance(x, Argument)]

            if invalid_types:
                raise HekrTypeError(variable='arguments list', expected=Argument, got=invalid_types)
        elif value is not None:
            raise HekrTypeError(variable='arguments list', expected=[list, type(None)], got=type(value))

        self.__arguments = value or []

    @property
    def response_command_id(self):
        return self.__response_command_id

    @response_command_id.setter
    def response_command_id(self, value):
        if value is not None:
            if self.frame_type is not None and self.frame_type != FrameType.SEND:
                raise HekrValueError(variable='response_command_id', expected='`None` due to `frame_type` not being set to `FrameType.SEND` (current value `%s`)' % self.frame_type, got=value)
            elif not isinstance(value, int):
                raise HekrTypeError(variable='response_command_id', expected=int, got=type(value))
            elif value < 0:
                raise HekrValueError(variable='response_command_id', expected='>= 0', got=value)
        self.__response_command_id = value


# -*- coding: utf-8 -*-
# pylint: disable=too-many-locals
"""Global constants for HekrAPI module"""

#: Default application ID used within requests via cloud and local transports
DEFAULT_APPLICATION_ID = 'hekr_python_connector'

#: First character of every presumably valid raw datagram
FRAME_START_IDENTIFICATION = 0x48

#: Default port to send UDP datagrams to
DEFAULT_DEVICE_PORT = 10000

#: Default amount of attempts a request should retry doing itself
DEFAULT_REQUEST_RETRIES = 2

#: Seconds count before attempting to perform next request iteration
DEFAULT_RETRY_DELAY = 1

#: Default host for controlling devices via websockets
DEFAULT_WEBSOCKET_HOST = 'fra-hub.hekreu.me'


#!/usr/bin/env python3
from distutils.core import setup
from subprocess import check_output
version = 'v0.0.7'
setup(
  name = 'hekrapi',
  packages = ['hekrapi', 'hekrapi.protocols'],
  version = version[1:],
  license='MIT',
  description = 'Python Hekr IoT API bindings',
  author = 'Alexander Ryazanov',
  author_email = 'alryaz@xavux.com',
  url = 'https://github.com/alryaz/hekrapi-python',
  download_url = 'https://github.com/alryaz/hekrapi-python/archive/' + version + '.tar.gz',
  keywords = ['Hekr', 'API', 'Wisen', 'Smart Devices', 'IoT'],
  install_requires=[
          'aiohttp',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)



""" Typing shortcuts for Hekr API """
from typing import List, Dict, Union, Any, AnyStr

DecodedDataType = Dict[str, Any]
DatagramType = AnyStr
CommandSearchIdentifier = Union[str, int]
AnyNumber = Union[int, float, None]


# -*- coding: utf-8 -*-
# pylint: disable=too-many-arguments
"""Protocol class module for Hekr API"""
from yaml import dump as yaml_dump, load as yaml_load, SafeLoader, SafeDumper
from json import dumps as json_dumps

from typing import Union, List, Dict
from collections import OrderedDict

from .types import (DecodedDataType, DatagramType, CommandSearchIdentifier)
from .command import Command, FrameType
from .exceptions import CommandNotFoundException, HekrTypeError, HekrValueError
from .datagram import encode, decode

def remove_none(obj):
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)((remove_none(k), remove_none(v))
        for k, v in obj.items() if k is not None and v is not None)
    else:
        return obj

class Protocol:
    """Protocol class definition for Hekr API"""

    def __init__(self, *args):
        """Protocol class constructor

        Arguments:
            *args {Command} -- Commands included in the protocol
        """
        self.commands = list(args)

    @property
    def commands(self) -> List[Command]:
        """Commands list setter

        Returns:
            List[Command] -- List of commands associated with the protocol
        """
        return self.__commands

    @commands.setter
    def commands(self, value:List[Command]) -> None:
        """Commands list setter

        Checks whether the value passed is of `list` type and that all elements of the list
        are of `Command` type.

        Arguments:
            value {[type]} -- [description]

        Raises:
            HekrTypeError: Raised when passed value is of a different type than `list`
            HekrTypeError: Raised when passed list contains elements of different type(s) than `list`
        """
        if not isinstance(value, list):
            raise HekrTypeError(variable='commands', expected=list, got=type(value))

        invalid_types = [type(x) for x in value
            if not isinstance(x, Command)]
        if invalid_types:
            raise HekrTypeError(variable='commands', expected=Command, got=invalid_types)

        self.__commands = value

    def __getitem__(self, key: CommandSearchIdentifier) -> Command:
        """Retrieves command with a square bracket accessor

        Arguments:
            key {CommandSearchIdentifier} -- Command ID/name

        Returns:
            Command -- Found command object
        """
        return self.get_command(key)

    def decode(self, raw: DatagramType,
               use_variable_names=False, filter_values=True) -> DecodedDataType:
        """Protocol-oriented datagram decoding

        Provides a wrapper for an out-of-class decoding function from `datagram` submodule.
        The wrapper passes all existing parameters alongside `self` as `protocol` argument.

        Arguments:
            raw {Union[bytearray, str]} -- Raw datagram string

        Keyword Arguments:
            use_variable_names {bool} -- Use variable names as keys with data (default: {False})
            filter_values {bool} -- Filter values (multiply/round) (default: {True})

        Returns:
            DecodedDataType -- Dictionary of result values
        """
        return decode(
            raw=raw,
            protocol=self,
            use_variable_names=use_variable_names,
            filter_values=filter_values
        )

    def encode(self,
               command: Union[int,
                              str,
                              Command],
               data: DecodedDataType,
               frame_number: int,
               use_variable_names=False,
               filter_values=True) -> DatagramType:
        """Protocol-oriented datagram encoding

        Provides a wrapper for an out-of-class encoding function from `datagram` submodule.
        The wrapper searches for a command within the protocol and passes all existing parameters
        alongside the result as `command` argument.

        Arguments:
            command {Union[int,str,Command]} -- Command identifier or object
            data {dict} -- Data dictionary
            frame_number {int} -- Frame number

        Keyword Arguments:
            use_variable_names {bool} -- Use variable names as keys with data (default: {False})
            filter_values {bool} -- Filter values (multiply/round) (default: {True})

        Returns:
            str -- Raw datagram string
        """
        return encode(
            command=self.get_command(command),
            data=data,
            frame_number=frame_number,
            use_variable_names=use_variable_names,
            filter_values=filter_values
        )

    def get_command_by_id(self, command_id: int) -> Command:
        """Get command by its ID

        Arguments:
            command_id {int} -- Command ID to search for

        Raises:
            CommandNotFoundException: Command not found by given ID

        Returns:
            Command -- Command object
        """
        for command in self.__commands:
            if command.command_id == command_id:
                return command

        raise CommandNotFoundException(command_id)

    def get_command_by_name(self, name: str) -> Command:
        """Get command by its name

        Arguments:
            name {str} -- Command name to search for

        Raises:
            CommandNotFoundException: Command not found by given name

        Returns:
            Command -- Command object
        """
        for command in self.__commands:
            if command.name == name:
                return command

        raise CommandNotFoundException(name)

    def get_command(self, command: Union[int, str, Command]) -> Command:
        """Get command by its ID/name (or cycle-return object)

        Arguments:
            command {Union[int, str, Command]} -- Command ID/name/object

        Raises:
            TypeError: Bad argument type provided

        Returns:
            Command -- Command object
        """
        if isinstance(command, Command):
            return command
        if isinstance(command, int):
            return self.get_command_by_id(command)
        if isinstance(command, str):
            return self.get_command_by_name(command)

        raise TypeError(
            "Argument 'command' (type %s) does not evaluate to any supported command type" %
            command)

    def get_definition(self, filter_empty=True):
        definition = {
            "commands": [
                {
                    "name": command.name,
                    "command_id": command.command_id,
                    "frame_type": command.frame_type.name.lower(),
                    "response_command_id": command.response_command_id,
                    "arguments": [
                        {
                            "name": argument.name,
                            "variable": argument.variable,
                            "byte_length": argument.byte_length,
                            "type_input": argument.type_input.__name__,
                            "type_output": argument.type_output.__name__,
                            "value_min": argument.value_min,
                            "value_max": argument.value_max,
                            "multiplier": argument.multiplier,
                            "decimals": argument.decimals,
                        }
                        for argument in command.arguments
                    ] if command.arguments else None
                }
                for command in self.commands
            ]
        }
        return remove_none(definition) if filter_empty else definition

    def print_definition(self, prefix: str = '', format:str='yaml'):
        """Prints protocol definition in YAML format

        Keyword Arguments:
            sort {bool} -- Sort commands by frame types (default: {False})
            prefix {str} -- What to prefix every line with (default: {''})
        """
        definition = self.get_definition()
        if format == 'yaml':
            output = yaml_dump(definition, Dumper=SafeDumper, default_flow_style=False, sort_keys=False)
        elif format == 'json':
            output = json_dumps(definition, ensure_ascii=False)

        print(output)
